// LM63.cpp : Defines the initialization routines for the DLL.
//
// created by Unwinder
/////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"

#include <float.h>
#include <afxdllx.h>

#include "LM63.h"
#include "RivaTunerExports.h"
#include "MonitoringSourceDesc.h"
/////////////////////////////////////////////////////////////////////////////
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
/////////////////////////////////////////////////////////////////////////////
static AFX_EXTENSION_MODULE LM63DLL = { NULL, NULL };
/////////////////////////////////////////////////////////////////////////////
GET_DEVICE_ID_PROC			g_pGetDeviceID			= NULL;
GET_I2C_BUS_NUM_PROC		g_pGetI2CBusNum			= NULL;
READ_I2C_PROC				g_pReadI2C				= NULL;

BYTE						g_bus					= 0;
BYTE						g_dev					= 0;

FLOAT						g_fltTemperatureRemote	= FLT_MAX;
FLOAT						g_fltTemperatureLocal	= FLT_MAX;
FLOAT						g_fltFanDutyCycle		= FLT_MAX;
FLOAT						g_fltFanSpeed			= FLT_MAX;

BOOL						g_bMisreadingFilter		= FALSE;
/////////////////////////////////////////////////////////////////////////////
extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	UNREFERENCED_PARAMETER(lpReserved);

	if (dwReason == DLL_PROCESS_ATTACH)
	{
		if (!AfxInitExtensionModule(LM63DLL, hInstance))
			return 0;

		new CDynLinkLibrary(LM63DLL);
	}
	else if (dwReason == DLL_PROCESS_DETACH)
	{
		AfxTermExtensionModule(LM63DLL);
	}
	return 1;
}
/////////////////////////////////////////////////////////////////////////////
// This helper function returns TRUE if the specified data should be treated
// as misread
//
// Background:
//
//Due to unknown reason (presumably hardware bug of R360's I2C controller)
//it can return errorneous zero data on successful I2C read transactions.
//We'll filter such values on these boards 
/////////////////////////////////////////////////////////////////////////////
BOOL IsMisreadData(BYTE data)
{
	return g_bMisreadingFilter && (data == 0);
}
/////////////////////////////////////////////////////////////////////////////
// This helper function is used to get fan speed or FLT_MAX if there is no
// fan connected to tachometer's input
//
// Note: 2 pulse per revolution tachometer is assumed in calcualtions
/////////////////////////////////////////////////////////////////////////////
FLOAT GetFanSpeed()
{
	FLOAT	result = FLT_MAX;

	WORD	tach;
	BYTE	data;

	if (g_pReadI2C(g_bus, g_dev, 0x46, &data) && !IsMisreadData(data))
		//read LSB
	{
		tach = data;

		if (g_pReadI2C(g_bus, g_dev, 0x47, &data) && !IsMisreadData(data))
			//read MSB and combine it with LSB
		{
			tach |= data<<8;

			if (tach)
				result = 5400000.0f / tach;

			if (tach == 0xffff)
				//handle fan stop case
				result = 0.0f;
		}
	}

	//perform misreading filtering

	if (result != FLT_MAX)
		g_fltFanSpeed = result;


	//return last correctly read value

	return g_fltFanSpeed;
}
/////////////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to get a number of
// data sources in this plugin
/////////////////////////////////////////////////////////////////////////////
LM63_API DWORD GetSourcesNum()
{
	return 4;
}
/////////////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to get descriptor
// for the specified data source
/////////////////////////////////////////////////////////////////////////////
LM63_API BOOL GetSourceDesc(DWORD dwIndex, LPMONITORING_SOURCE_DESC pDesc)
{
	//get host application module handle

	HMODULE hHost = GetModuleHandle(NULL);
	if (!hHost)
		return FALSE;

	if (!g_dev)
		//small performance optimization: once the sensor is detected, we'll
		//precache its' location and will no longer perform I2C scanning
	{
		//get ptr to host application ID routine and return error if the host doesn't 
		//export it

		//we'll use this condition to verify if we're under RivaTuner v2.0 RC 15.7
		//(older versions are not supported) or any other compatible host

		GET_HOST_APP_ID_PROC pGetHostAppID = (GET_HOST_APP_ID_PROC)GetProcAddress(hHost, "GetHostAppID");
		if (!pGetHostAppID)
			return FALSE;

		//get ptrs to required plugin API functions

		g_pGetDeviceID		= (GET_DEVICE_ID_PROC)GetProcAddress(hHost, "GetDeviceID");
		if (!g_pGetDeviceID)
			return FALSE;

		g_pGetI2CBusNum		= (GET_I2C_BUS_NUM_PROC)GetProcAddress(hHost, "GetI2CBusNum");
		if (!g_pGetI2CBusNum)
			return FALSE;

		g_pReadI2C			= (READ_I2C_PROC)GetProcAddress(hHost, "ReadI2C");
		if (!g_pReadI2C)
			return FALSE;

		//once we've got all function ptrs, we can proceed with I2C bus scanning

		//first, we'll get number of I2C buses for the selected display device

		BYTE busNum		= g_pGetI2CBusNum();
		if (!busNum)
			return FALSE;

		//scan all I2C buses and check LM63-compatible sensor presence	

		for (BYTE bus=0; bus<busNum; bus++)
		{
			BYTE data;

			if (g_pReadI2C(bus, 0x18, 0xFE, &data) && (data == 0x01))
					//check LM64 manufacturer ID
			{
				if (g_pReadI2C(bus, 0x18, 0xFF, &data) && (data == 0x51))
					//check LM64 revision ID
				{
					//we've found LM64 and can precache its' location

					g_bus = bus;
					g_dev = 0x18;
					break;
				}
			}

			if (g_pReadI2C(bus, 0x4E, 0xFE, &data) && (data == 0x01))
					//check LM64 manufacturer ID
			{
				if (g_pReadI2C(bus, 0x4E, 0xFF, &data) && (data == 0x51))
					//check LM64 revision ID
				{
					//we've found LM64 and can precache its' location

					g_bus = bus;
					g_dev = 0x4E;
					break;
				}
			}

			if (g_pReadI2C(bus, 0x4C, 0xFE, &data)) 
				if (data == 0x01)
					//check LM63 manufacturer ID
				{
					if (g_pReadI2C(bus, 0x4C, 0xFF, &data) && (data == 0x41))
						//check LM63/LM64 revision ID
					{
						//we've found LM63 and can precache its' location

						g_bus = bus;
						g_dev = 0x4C;
						break;
					}
				}
				else
					if (data == 0x23)
						//check F75363S manufacturer ID
					{
						if (g_pReadI2C(bus, 0x4C, 0xFF, &data) && ((data == 0x11) || (data == 0x20)))
							//check F75363S revision ID (both retail and engineering revisions)
						{
							//we've found F75363S and can precache its' location

							g_bus = bus;
							g_dev = 0x4C;
							break;
						}
					}
		}
	}

	if (!g_dev)
		//return FALSE if we've not detected LM63-compatible sensor
		return FALSE;

	//now we've to setup additional misreading filter for 9800XT series

	//due to unknown reason (presumably hardware bug of R360's I2C controller)
	//it can return errorneous zero data on successful I2C read transactions.

	//we'll filter such values on these boards

	CString strDeviceID = g_pGetDeviceID();	strDeviceID.MakeUpper();

	DWORD	vendorID, deviceID;
	DWORD	subsysID;
	DWORD	revision;

	sscanf(strDeviceID,"PCI\\VEN_%04x&DEV_%04x&SUBSYS_%08x&REV_%02x", &vendorID, &deviceID, &subsysID, &revision);

	if ((vendorID == 0x1002) && ((deviceID & 0xfff0) == 0x4E40))
		//it is R350/R360, enable misreading filtering
		g_bMisreadingFilter = TRUE;

	switch (dwIndex)
		//now the sensor's location is detected,
		//so we can return requested information to RivaTuner
	{
	case 0:
		//fill core temperature source descriptor

		strcpy(pDesc->szName	, "Core temperature");
		strcpy(pDesc->szDim		, "C");

		pDesc->fltMaxLimit	= 200.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Remote temperature reported by LM63-compatible sensor ");

		return TRUE;

	case 1:
		//fill ambient temperature source descriptor

		strcpy(pDesc->szName	, "Ambient temperature");
		strcpy(pDesc->szDim		, "C");

		pDesc->fltMaxLimit	= 200.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Local temperature reported by LM63-compatible sensor ");

		return TRUE;

	case 2:
		//fill fan duty cycle source descriptor

		strcpy(pDesc->szName	, "Fan duty cycle");
		strcpy(pDesc->szDim		, "%");

		pDesc->fltMaxLimit	= 100.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Fan duty cycle reported by LM63-compatible sensor ");

		return TRUE;

	case 3:
		//check if fan is connected to tachometer and fill fan speed source
		//descriptor if necessary

		if (GetFanSpeed() == FLT_MAX)
			return FALSE;

		strcpy(pDesc->szName	, "Fan speed");
		strcpy(pDesc->szDim		, "RPM");

		pDesc->fltMaxLimit	= 10000.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Fan speed reported by LM63-compatible sensor ");

		return TRUE;
	}

	return FALSE;
}
/////////////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to poll data sources
/////////////////////////////////////////////////////////////////////////////
LM63_API FLOAT GetSourceData(DWORD dwIndex)
{
	FLOAT	result	= FLT_MAX;	
	SHORT	temp;
	BYTE	data;

	switch (dwIndex)
	{
	case 0:
		//core temperature

		if (g_pReadI2C(g_bus, g_dev, 1, &data) && !IsMisreadData(data))
			//read MSB
		{
			temp = data<<8;

			if (g_pReadI2C(g_bus, g_dev, 0x10, &data) && !IsMisreadData(data))
				//read LSB and combine it with MSB
				temp |= data;

			result = temp / 256.0f;
		}

		//perform misreading filtering

		if (result != FLT_MAX)
			g_fltTemperatureRemote = result;

		//return last correctly read value

		return g_fltTemperatureRemote;

	case 1:
		//ambient temperature

		if (g_pReadI2C(g_bus, g_dev, 0, &data) && !IsMisreadData(data))
			//read MSB
		{
			temp = data<<8;

			if (g_pReadI2C(g_bus, g_dev, 0x11, &data) && !IsMisreadData(data))
				//read LSB and combine it with MSB
				temp |= data;

			result = temp / 256.0f;
		}

		//perform misreading filtering

		if (result != FLT_MAX)
			g_fltTemperatureLocal = result;

		//return last correctly read value

		return g_fltTemperatureLocal;

	case 2:
		//fan duty cycle

		{
			BOOL bPolarity = FALSE;

			if (g_pReadI2C(g_bus, g_dev, 0x4A, &data) && !IsMisreadData(data))
				//read PWM configuration
			{
				if (data & 0x10)
					bPolarity = TRUE;

				if (g_pReadI2C(g_bus, g_dev, 0x4D, &data) && !IsMisreadData(data))
					//read PWM frequency
				{
					FLOAT freq = (FLOAT)(data & 0x1f);

					if (g_pReadI2C(g_bus, g_dev, 0x4C, &data) && !IsMisreadData(data))
						//read PWM value and calculate duty cycle
					{

						FLOAT val = (FLOAT)(data & 0x3f);
				
						result = val * 100.0f / (2.0f * freq);

						if (result > 100.0f)
							result = 100.0f;

						if (!bPolarity)
							result = 100.0f - result;
					}

				}
			}
		}

		//perform misreading filtering

		if (result != FLT_MAX)
			g_fltFanDutyCycle = result;

		//return last correctly read value

		return g_fltFanDutyCycle;

	case 3:
		//fan speed

		return GetFanSpeed();

	}

	return result;
}
/////////////////////////////////////////////////////////////////////////////
