// NVThermalDiode.cpp : Defines the initialization routines for the DLL.
//
// created by Unwinder
/////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"

#include <float.h>
#include <afxdllx.h>

#include "NVThermalDiode.h"
#include "RivaTunerExports.h"
#include "MonitoringSourceDesc.h"
/////////////////////////////////////////////////////////////////////////////
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
/////////////////////////////////////////////////////////////////////////////
static AFX_EXTENSION_MODULE NVThermalDiodeDLL = { NULL, NULL };
/////////////////////////////////////////////////////////////////////////////
GET_DEVICE_ID_PROC			g_pGetDeviceID				= NULL;
READ_REGISTER_ULONG_PROC	g_pReadRegisterUlong		= NULL;
WRITE_REGISTER_ULONG_PROC	g_pWriteRegisterUlong		= NULL;
ENABLE_READ_BIOS_PROC		g_pEnableReadBIOS			= NULL;
READ_BIOS_PROC				g_pReadBIOS					= NULL;

LONG						g_dwDiodeGainMul			= 0;
LONG						g_dwDiodeGainDiv			= 0;
LONG						g_dwDiodeOffsetMul			= 0;
LONG						g_dwDiodeOffsetDiv			= 0;

LONG						g_dwTemperatureCompensation	= 0;
LONG						g_dwTemperatureThreshold	= 125;
DWORD						g_dwMask					= 0;

DWORD						g_dwCoreFamily				= 0;
/////////////////////////////////////////////////////////////////////////////
extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	UNREFERENCED_PARAMETER(lpReserved);

	if (dwReason == DLL_PROCESS_ATTACH)
	{
		if (!AfxInitExtensionModule(NVThermalDiodeDLL, hInstance))
			return 0;

		new CDynLinkLibrary(NVThermalDiodeDLL);
	}
	else if (dwReason == DLL_PROCESS_DETACH)
	{
		AfxTermExtensionModule(NVThermalDiodeDLL);
	}
	return 1;
}
//////////////////////////////////////////////////////////////////////
// This helper function is used to find specified pattern in VGA BIOS
// image
//////////////////////////////////////////////////////////////////////
LPBYTE FindPattern(LPBYTE lpImage, LPBYTE lpPattern, LPBYTE lpMask, DWORD dwSize, DWORD dwOffset)
{
	DWORD dwImageSize = lpImage[2] * 512;

	while (dwOffset < dwImageSize - dwSize)
	{
		BOOL bMatched = TRUE;

		for (DWORD dwIndex=0; dwIndex<dwSize; dwIndex++)
		{
			BYTE src = lpImage[dwOffset + dwIndex];
			BYTE dst = lpPattern[dwIndex];

			if (lpMask)
			{
				src &= lpMask[dwIndex];
				dst &= lpMask[dwIndex];
			}

			if (src != dst)
			{
				bMatched = FALSE;
				break;
			}
		}
		
		if (bMatched)
			return lpImage + dwOffset;

		dwOffset++;
	}

	return NULL;
}
//////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to get a number of
// data sources in this plugin
//////////////////////////////////////////////////////////////////////
NVTHERMALDIODE_API DWORD GetSourcesNum()
{
	return 1;
}
//////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to get descriptor
// for the specified data source
//////////////////////////////////////////////////////////////////////
NVTHERMALDIODE_API BOOL GetSourceDesc(DWORD dwIndex, LPMONITORING_SOURCE_DESC pDesc)
{
	//get host application module handle

	HMODULE hHost = GetModuleHandle(NULL);
	if (!hHost)
		return FALSE;

	//get ptr to host application ID routine and return error if the host doesn't 
	//export it

	//we'll use this condition to verify if we're under RivaTuner v2.0 RC 15.7
	//(older versions are not supported) or any other compatible host

	GET_HOST_APP_ID_PROC pGetHostAppID = (GET_HOST_APP_ID_PROC)GetProcAddress(hHost, "GetHostAppID");
	if (!pGetHostAppID)
		return FALSE;

	//get ptrs to required plugin API functions

	g_pGetDeviceID			= (GET_DEVICE_ID_PROC)GetProcAddress(hHost, "GetDeviceID");
	if (!g_pGetDeviceID)
		return FALSE;

	g_pReadRegisterUlong	= (READ_REGISTER_ULONG_PROC)GetProcAddress(hHost, "ReadRegisterUlong");
	if (!g_pReadRegisterUlong)
		return FALSE;

	g_pWriteRegisterUlong	= (WRITE_REGISTER_ULONG_PROC)GetProcAddress(hHost, "WriteRegisterUlong");
	if (!g_pWriteRegisterUlong)
		return FALSE;

	g_pEnableReadBIOS		= (ENABLE_READ_BIOS_PROC)GetProcAddress(hHost, "EnableReadBIOS");
	if (!g_pEnableReadBIOS)
		return FALSE;

	g_pReadBIOS	= (READ_BIOS_PROC)GetProcAddress(hHost, "ReadBIOS");
	if (!g_pReadBIOS)
		return FALSE;

	//once we've got all function ptrs, we can proceed with hardware checking

	//first, we've to check if it is NVIDIA display adapter

	CString strDeviceID = g_pGetDeviceID();	strDeviceID.MakeUpper();

	DWORD	vendorID, deviceID;
	DWORD	subsysID;
	DWORD	revision;

	sscanf(strDeviceID,"PCI\\VEN_%04x&DEV_%04x&SUBSYS_%08x&REV_%02x", &vendorID, &deviceID, &subsysID, &revision);

	if (vendorID != 0x10DE)
		//return error if display adapter is not manufactured by NVIDIA
		return FALSE;

	//second, we've to check if GPU has on-die thermal diode
	
	g_dwCoreFamily = (g_pReadRegisterUlong(0)>>20) & 0xff;
		//get GPU core family ID

	//init default thermal diode calibration parameters for thermal diode 
	//capable GPUs

	switch (g_dwCoreFamily)
	{
	case 0x43:
		//init default thermal diode calibration params for NV43
		g_dwDiodeGainMul	= 792;
		g_dwDiodeGainDiv	= 1000;
		g_dwDiodeOffsetMul	= 32060;
		g_dwDiodeOffsetDiv	= 1000;
		g_dwMask			= 0xff;
		break;
	case 0x44:
	case 0x4A:
	case 0x46:
	case 0x47:
		//init default thermal diode calibration params for NV44+
		g_dwDiodeGainMul	= 780;
		g_dwDiodeGainDiv	= 1000;
		g_dwDiodeOffsetMul	= 27839;
		g_dwDiodeOffsetDiv	= 1000;
		g_dwMask			= 0xff;
		break;
	case 0x49:
	case 0x4B:
		//init default thermal diode calibration parms for G71+
		g_dwDiodeGainMul	= 450;
		g_dwDiodeGainDiv	= 10000;
		g_dwDiodeOffsetMul	= -23517;
		g_dwDiodeOffsetDiv	= 100;
		g_dwMask			= 0x1fff;
		break;
	default: 
		//return error if GPU has no on-die thermal diode
		return FALSE;
	}

	//third, we've to override default thermal diode calibration parameters with
	//BIOS defined ones from BIOS thermal table if possible

	if (g_pEnableReadBIOS(TRUE))
		//enable access to BIOS before actually reading it
	{
		if ((g_pReadBIOS(0) == 0x55) &&
			(g_pReadBIOS(1) == 0xAA))
			//check if VGA BIOS access actually works by verifying standard
			//VGA BIOS signature
		{
			DWORD dwSize	= g_pReadBIOS(2) * 512;
				//get ROM size

			if (dwSize)
			{
				LPBYTE lpImage	= new BYTE[dwSize];
					//allocate buffer for temporary VGA BIOS image storage

				for (DWORD dwIndex=0; dwIndex<dwSize; dwIndex++)
					lpImage[dwIndex] = g_pReadBIOS(dwIndex);
					//copy BIOS to allocated buffer

				const BYTE signature[] = { 0xFF, 0xB8, 0x42, 0x49, 0x54, 0x00 };
					//define BIT table signature, we'll use it to find BIT in BIOS image

				LPBYTE lpBIT = FindPattern(lpImage, (LPBYTE)signature, NULL, sizeof(signature), 0);
					//try to find BIT table in BIOS image by signature

				if (lpBIT)
				{
					for (DWORD dwToken=0; dwToken<lpBIT[10]; dwToken++)
						//loop though all BIT tokens 
					{
						LPBYTE lpToken = lpBIT + lpBIT[8] + lpBIT[9] * dwToken;
							//get ptr to current BIT token

						if (lpToken[0] == 0x50)
							//we've found performance area token and can now
							//get offset of performance area
						{
							WORD offPerfArea = *(LPWORD)(lpToken + 0x04);

							if (offPerfArea)
								//once performance area offset is determined, we
								//can read offset of thermal table
							{
								WORD offThermTable = *(LPWORD)(lpImage + offPerfArea + 0x0C);

								if (offThermTable)
									//now we've determined thermal table offset and
									//can scan it to read thermal diode calibration
									//parameters, temperature compensation and
									//threshold
								{
									LPBYTE lpThermTable = lpImage + offThermTable;

									DWORD	dwTemperatureChannel = 0x0f;
										//init current temperature channel identifier
		
									for (DWORD dwEntry=0; dwEntry<lpThermTable[3]; dwEntry++)
										//loop through all thermal table entries
									{
										LPBYTE	lpEntry = lpThermTable + lpThermTable[1] + lpThermTable[2] * dwEntry;
											// get ptr to current thermal table entry
										WORD	value	= *(LPWORD)(lpEntry + 1);
											//get thermal table entry value
						
										switch (lpEntry[0])
											//we'll only process entries calibrating temperature readings and defining
											//core temperature threshold
										{
										case 0x00:	//temperature channel 
											dwTemperatureChannel = value & 0x0f;
											break;
										case 0x01:	//temperature compensation
											if ((value & 0x8F) == 0)
												g_dwTemperatureCompensation = (value>>9) & 0x7f;
											break;
										case 0x04:	//temperature threshold
											if (dwTemperatureChannel == 0)
												//process threshold for core temperature channel only
												g_dwTemperatureThreshold = (value>>4) & 0x1ff;
											break;
										case 0x10:	//thermal diode offset multiplier
											g_dwDiodeOffsetMul = (SHORT)value;
											break;
										case 0x11:	//thermal diode offset divider
											g_dwDiodeOffsetDiv = (SHORT)value;
											break;
										case 0x12:	//thermal diode gain multiplier
											g_dwDiodeGainMul = (SHORT)value;
											break;
										case 0x13:	//thermal diode gain divider
											g_dwDiodeGainDiv = (SHORT)value;
											break;
										}
									}
								}
							}
						}
					}
				}

				delete [] lpImage;
					//we no longer need VGA BIOS image and can deallocate it now
			}
		}

		g_pEnableReadBIOS(FALSE);
			//don't forget to disable access to BIOS when we no longer need it
	}


	if (!g_dwDiodeGainMul || !g_dwDiodeGainDiv || !g_dwDiodeOffsetDiv || !g_dwTemperatureThreshold)
		//return error if thermal table stored in BIOS is corrupted
		return FALSE;

	switch (dwIndex)
		//now thermal diode capable GPU is detected
		//so we can return requested information to RivaTuner
	{
	case 0:
		//fill core temperature source descriptor

		strcpy(pDesc->szName	, "Core temperature");
		strcpy(pDesc->szDim		, "C");

		pDesc->fltMaxLimit	= 200.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Temperature reported by GPU on-die thermal diode ");

		return TRUE;
	}

	return FALSE;
}
/////////////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to poll data sources
/////////////////////////////////////////////////////////////////////////////
NVTHERMALDIODE_API FLOAT GetSourceData(DWORD dwIndex)
{
	FLOAT	result	= FLT_MAX;	

	switch (dwIndex)
	{
	case 0:	//core temperature

		LONG temp = g_pReadRegisterUlong(0x15b4) & g_dwMask;
			//get raw thermal diode readings

		if (!temp)
			//check if diode is actually enabled, otherwise enable it by programming
			//temperature threshold
			g_pWriteRegisterUlong(0x15b0, 0x10000000 | ((g_dwTemperatureThreshold - g_dwTemperatureCompensation - g_dwDiodeOffsetMul / g_dwDiodeOffsetDiv) * g_dwDiodeGainDiv / g_dwDiodeGainMul));
		else
			result = (FLOAT)(temp * g_dwDiodeGainMul / g_dwDiodeGainDiv + g_dwDiodeOffsetMul / g_dwDiodeOffsetDiv + g_dwTemperatureCompensation);
				//calibrate thermal diode readings
		break;
	}

	return result;
}
/////////////////////////////////////////////////////////////////////////////
