// VT1103.cpp : Defines the initialization routines for the DLL.
//
// created by Unwinder
/////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"

#include <float.h>
#include <afxdllx.h>
#include <shlwapi.h>

#include "VT1103.h"
#include "RivaTunerExports.h"
#include "MonitoringSourceDesc.h"
/////////////////////////////////////////////////////////////////////////////
#ifdef _DEBUG
#define new DEBUG_NEW
#undef THIS_FILE
static char THIS_FILE[] = __FILE__;
#endif
/////////////////////////////////////////////////////////////////////////////
static AFX_EXTENSION_MODULE VT1103DLL = { NULL, NULL };
/////////////////////////////////////////////////////////////////////////////
GET_DEVICE_ID_PROC			g_pGetDeviceID			= NULL;
GET_I2C_BUS_NUM_PROC		g_pGetI2CBusNum			= NULL;
READ_I2C_PROC				g_pReadI2C				= NULL;

BYTE						g_bus					= 0;
BYTE						g_dev					= 0;

FLOAT						g_fltVoltage			= FLT_MAX;
FLOAT						g_fltTemperatureLocal	= FLT_MAX;
FLOAT						g_fltAverageCurrent		= FLT_MAX;

FLOAT						g_voltageTable[0x40]	= { 0 };

HINSTANCE					g_hModule				= 0;
/////////////////////////////////////////////////////////////////////////////
extern "C" int APIENTRY
DllMain(HINSTANCE hInstance, DWORD dwReason, LPVOID lpReserved)
{
	UNREFERENCED_PARAMETER(lpReserved);

	if (dwReason == DLL_PROCESS_ATTACH)
	{
		if (!AfxInitExtensionModule(VT1103DLL, hInstance))
			return 0;

		new CDynLinkLibrary(VT1103DLL);

		g_hModule = hInstance;
	}
	else if (dwReason == DLL_PROCESS_DETACH)
	{
		AfxTermExtensionModule(VT1103DLL);
	}
	return 1;
}
//////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to get a number of
// data sources in this plugin
//////////////////////////////////////////////////////////////////////
VT1103_API DWORD GetSourcesNum()
{
	return 3;
}
//////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to get descriptor
// for the specified data source
//////////////////////////////////////////////////////////////////////
VT1103_API BOOL GetSourceDesc(DWORD dwIndex, LPMONITORING_SOURCE_DESC pDesc)
{
	//get host application module handle

	HMODULE hHost = GetModuleHandle(NULL);
	if (!hHost)
		return FALSE;

	if (!g_dev)
		//small performance optimization: once the sensor is detected, we'll
		//precache its' location and will no longer perform I2C scanning
	{
		//get ptr to host application ID routine and return error if the host doesn't 
		//export it

		//we'll use this condition to verify if we're under RivaTuner v2.0 RC 15.7
		//(older versions are not supported) or any other compatible host

		GET_HOST_APP_ID_PROC pGetHostAppID = (GET_HOST_APP_ID_PROC)GetProcAddress(hHost, "GetHostAppID");
		if (!pGetHostAppID)
			return FALSE;

		//get ptrs to required plugin API functions

		g_pGetDeviceID		= (GET_DEVICE_ID_PROC)GetProcAddress(hHost, "GetDeviceID");
		if (!g_pGetDeviceID)
			return FALSE;

		g_pGetI2CBusNum		= (GET_I2C_BUS_NUM_PROC)GetProcAddress(hHost, "GetI2CBusNum");
		if (!g_pGetI2CBusNum)
			return FALSE;

		g_pReadI2C			= (READ_I2C_PROC)GetProcAddress(hHost, "ReadI2C");
		if (!g_pReadI2C)
			return FALSE;

		//once we've got all function ptrs, we can proceed with I2C bus scanning

		//first, we'll get number of I2C buses for the selected display device

		BYTE busNum		= g_pGetI2CBusNum();
		if (!busNum)
			return FALSE;

		//scan all I2C buses and check VT1103 presence	

		for (BYTE bus=0; bus<busNum; bus++)
		{
			BYTE data;

			if (g_pReadI2C(bus, 0x70, 0x1A, &data) && (data == 0x07))
			{
				g_bus = bus;
				g_dev = 0x70;
				break;
			}
		}

		//read voltage table from configuration file

		char szCfgPath[MAX_PATH];
		GetModuleFileName(g_hModule, szCfgPath, MAX_PATH);
		
		PathRenameExtension(szCfgPath, ".cfg");
			//get fully qualified path to .cfg

		for (DWORD dwVID=0; dwVID<0x40; dwVID++)
		{
			char szVID[MAX_PATH];
			sprintf(szVID, "%02X", dwVID);
				//format VID value string

			GetPrivateProfileString("VID", szVID, "", szVID, MAX_PATH, szCfgPath);
				//read voltage mapping for current VID from .cfg

			sscanf(szVID, "%f", g_voltageTable + dwVID);
				//convert voltage mapping from string to float
		}
	}

	if (!g_dev)
		//return FALSE if we've not detected VT1103 voltage regulator
		return FALSE;

	switch (dwIndex)
		//now the sensor's location is detected,
		//so we can return requested information to RivaTuner
	{
	case 0:
		//fill voltage regulator output source descriptor

		strcpy(pDesc->szName	, "Voltage regulator output");
		strcpy(pDesc->szDim		, "V");

		pDesc->fltMaxLimit	= 2.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Output voltage reported by VT1103 voltage regulator ");

		return TRUE;

	case 1:
		//fill voltage regulator temperature source descriptor

		strcpy(pDesc->szName	, "Voltage regulator temperature");
		strcpy(pDesc->szDim		, "C");

		pDesc->fltMaxLimit	= 200.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Local temperature reported by VT1103 voltage regulator ");

		return TRUE;

	case 2:
		//fill voltage regulator average current

		strcpy(pDesc->szName	, "Voltage regulator current");
		strcpy(pDesc->szDim		, "A");

		pDesc->fltMaxLimit	= 50.0f;
		pDesc->fltMinLimit	= 0.0f;
		pDesc->fltGridDim	= 10.0f;

		strcpy(pDesc->szDesc, " Average current reported by VT1103 voltage regulator ");

		return TRUE;
	}

	return FALSE;
}
/////////////////////////////////////////////////////////////////////////////
// This exported function is called by RivaTuner to poll data sources
/////////////////////////////////////////////////////////////////////////////
VT1103_API FLOAT GetSourceData(DWORD dwIndex)
{
	FLOAT	result	= FLT_MAX;	
	BYTE	temp;
	BYTE	data;

	switch (dwIndex)
	{
	case 0:
		//voltage regulator output

		if (g_pReadI2C(g_bus, g_dev, 0x1C, &data))
			result = g_voltageTable[data & 0x3f];

		//perform misreading filtering

		if (result != FLT_MAX)
			g_fltVoltage = result;

		//return last correctly read value

		return g_fltVoltage;

	case 1:
		//voltage regulator temperature

		if (g_pReadI2C(g_bus, g_dev, 0x17, &data))
			result = (char)data;

		//perform misreading filtering

		if (result != FLT_MAX)
			g_fltTemperatureLocal = result;

		//return last correctly read value

		return g_fltTemperatureLocal;

	case 2:
		//voltage regulator average current

		if (g_pReadI2C(g_bus, g_dev, 0x14, &data) &&
			g_pReadI2C(g_bus, g_dev, 0x15, &data))
		{
			temp = data>>1;
			if (!temp)
				temp = 1;

			if (g_pReadI2C(g_bus, g_dev, 0x11, &data) &&
				g_pReadI2C(g_bus, g_dev, 0x12, &data))
				result = 60.0f * (data>>1) / temp;
		}

		//perform misreading filtering

		if (result != FLT_MAX)
			g_fltAverageCurrent = result;

		//return last correctly read value

		return g_fltAverageCurrent;
	}

	return result;
}
/////////////////////////////////////////////////////////////////////////////
